/*
 * Copyright 2021 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.tomee;

import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.catalina.TomcatControllerHandler;
import org.apache.catalina.startup.Catalina;
import org.apache.geronimo.transaction.manager.TransactionTimer;
import org.apache.openejb.OpenEJBCoreControllerHandler;
import org.apache.openejb.loader.Files.DeleteThread;
import org.apache.tomcat.util.scan.TomcatUtilScanControllerHandler;
import org.apache.tomcat.vestige.launcher.TomcatVestigeLauncher;
import org.apache.tomcat.vestige.launcher.VestigeTomcatController;
import org.apache.tomcat.vestige.launcher.VestigeWar;
import org.apache.tomee.catalina.CatalinaControllerHandler;
import org.apache.tomee.catalina.TomEEClassLoaderEnricher;

import fr.gaellalire.vestige.spi.resolver.maven.VestigeMavenResolver;

/**
 * @author Gael Lalire
 */
public class TomEEVestigeLauncher extends TomcatVestigeLauncher implements Runnable {

    private static final Logger LOGGER = Logger.getLogger(TomEEVestigeLauncher.class.getName());

    private VestigeMavenResolver mavenResolver;

    public void setMavenResolver(final VestigeMavenResolver mavenResolver) {
        this.mavenResolver = mavenResolver;
    }

    private TomEEVestigeSystem vestigeSystem;

    public void setVestigeSystem(final TomEEVestigeSystem vestigeSystem) {
        super.setVestigeSystem(vestigeSystem);
        this.vestigeSystem = vestigeSystem;
    }

    public TomEEVestigeLauncher(final File conf, final File data) {
        super(conf, data);
        System.setProperty("tomee.serialization.class.blacklist", "-");

        System.setProperty(TomEEClassLoaderEnricher.TOMEE_WEBAPP_CLASSLOADER_ENRICHMENT_SKIP, "true");
    }

    private volatile boolean started = false;

    @SuppressWarnings("deprecation")
    public void run() {
        VestigeTomEEController vestigeTomeeController = new VestigeTomEEController(vestigeSystem);
        OpenEJBCoreControllerHandler.setOpenEJBCoreController(vestigeTomeeController);
        CatalinaControllerHandler.setCatalinaController(vestigeTomeeController);
        File unpackDir = new File(getData(), "unpackDir");
        unpackDir.mkdirs();
        VestigeEar.init(mavenResolver);
        VestigeTomcatController vestigeTomcatController = new VestigeTomcatController();
        try {
            TomcatControllerHandler.setTomcatController(vestigeTomcatController);
            TomcatUtilScanControllerHandler.setTomcatController(vestigeTomcatController);
            VestigeWar.init(this);

            final Catalina catalina = new Catalina() {
                @Override
                protected void initStreams() {
                    // system.out & system.err are already hooked
                }
            };
            catalina.setUseShutdownHook(false);
            catalina.setParentClassLoader(TomEEVestigeLauncher.class.getClassLoader());

            // clBinding for initial context
            Thread.currentThread().setContextClassLoader(Catalina.class.getClassLoader());

            Thread launcherThread = new Thread() {
                @Override
                public void run() {
                    catalina.start();
                    started = true;
                }
            };
            launcherThread.start();
            try {
                synchronized (this) {
                    wait();
                }
            } catch (InterruptedException e) {
                while (true) {
                    try {
                        launcherThread.join();
                        break;
                    } catch (InterruptedException e1) {
                        LOGGER.log(Level.FINE, "Ignore interrupt", e1);
                    }
                }
                if (started) {
                    catalina.stop();
                }

                DeleteThread deleteThread = new DeleteThread();
                deleteThread.start();

                Thread currentThread = Thread.currentThread();
                ThreadGroup threadGroup = currentThread.getThreadGroup();
                int activeCount = threadGroup.activeCount();
                while (activeCount != 1) {
                    Thread[] list = new Thread[activeCount];
                    int enumerate = threadGroup.enumerate(list);
                    for (int i = 0; i < enumerate; i++) {
                        Thread t = list[i];
                        if (t == currentThread) {
                            continue;
                        }
                        t.interrupt();
                    }
                    for (int i = 0; i < enumerate; i++) {
                        Thread t = list[i];
                        if (t == currentThread) {
                            continue;
                        }

                        // TransactionTimer.CurrentTime hack
                        Class<?> currentTimeClass;
                        try {
                            currentTimeClass = Class.forName(TransactionTimer.class.getName() + "$CurrentTime");
                            if (currentTimeClass.isInstance(t)) {
                                // this class has no shutdown mecanism, and is
                                // catching InterruptedException
                                t.stop();
                                LOGGER.log(Level.INFO, "stop called ");
                            }
                        } catch (ClassNotFoundException e2) {
                            LOGGER.log(Level.FINE, "ClassNotFoundException", e2);
                        }

                        try {
                            t.join();
                        } catch (InterruptedException e1) {
                            LOGGER.log(Level.FINE, "Interrupted", e1);
                            break;
                        }
                    }
                    activeCount = threadGroup.activeCount();
                }

            } finally {
                started = false;
            }
        } finally {
            try {
                vestigeTomcatController.close();
            } catch (IOException e) {
                // ignore
            }
        }

    }

}
